/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "assert/assert.h"
#include "util/memory.h"
#include "entity.h"
#include "rpg/world/map.h"
#include "util/math.h"

entitycallbacks_t ENTITY_CALLBACKS[ENTITY_TYPE_COUNT] = {
    [ENTITY_TYPE_NULL] = { 0 },
    
    [ENTITY_TYPE_PLAYER] = {
      .init = playerInit,
      .update = playerUpdate,
      .interact = NULL,
    },

    [ENTITY_TYPE_NPC] = {
      .init = npcInit,
      .update = npcUpdate,
      .interact = npcInteract,
    }
};

void entityInit(entity_t *entity, const entitytype_t type) {
  assertNotNull(entity, "Entity is NULL");
  assertTrue(type < ENTITY_TYPE_COUNT, "Invalid entity type");
  assertNotNull(ENTITY_CALLBACKS[type].init, "Entity type has no init");

  memoryZero(entity, sizeof(entity_t));
  entity->type = type;
  ENTITY_CALLBACKS[type].init(entity);
}

void entityUpdate(entity_t *entity) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertNotNull(
    ENTITY_CALLBACKS[entity->type].update,
    "Entity type has no update"
  );

  // Handle subpixel movement
  if(entity->subX < 0) {
    entity->subX++;
  } else if(entity->subY < 0) {
    entity->subY++;
  } else if(entity->subX > 0) {
    entity->subX--;
  } else if(entity->subY > 0) {
    entity->subY--;
  }

  // Entity-Type handling
  ENTITY_CALLBACKS[entity->type].update(entity);
}

void entityTurn(entity_t *entity, const entitydir_t dir) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertFalse(entityIsWalking(entity), "Entity is currently walking");
  entity->dir = dir;
}

void entityDirGetPosition(
  const entitydir_t dir,
  const uint8_t distance,
  int8_t *outX,
  int8_t *outY
) {
  assertTrue(dir < ENTITY_DIR_COUNT, "Invalid entity direction");
  assertNotNull(outX, "Output X cannot be NULL");
  assertNotNull(outY, "Output Y cannot be NULL");

  switch(dir) {
    case ENTITY_DIR_UP:
      *outX = 0;
      *outY = -distance;
      break;
    case ENTITY_DIR_DOWN:
      *outX = 0;
      *outY = distance;
      break;
    case ENTITY_DIR_LEFT:
      *outX = -distance;
      *outY = 0;
      break;
    case ENTITY_DIR_RIGHT:
      *outX = distance;
      *outY = 0;
      break;
    default:
      assertUnreachable("Invalid entity direction");
  }
}

entitydir_t entityGetLookDirection(
  const entity_t *entity,
  const uint8_t x,
  const uint8_t y
) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertTrue(x < MAP.width, "X coordinate out of bounds");
  assertTrue(y < MAP.height, "Y coordinate out of bounds");

  int8_t dX = x - entity->x;
  int8_t dY = y - entity->y;

  // More horizontal movement or more vertical movement?
  if(mathAbsI8(dX) > mathAbsI8(dY)) {
    // More horizontal movement
    if(dX < 0) return ENTITY_DIR_LEFT;
    return ENTITY_DIR_RIGHT;
  }

  if(dY < 0) {
    return ENTITY_DIR_UP;
  }
  return ENTITY_DIR_DOWN;
}

void entityWalk(entity_t *entity) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertFalse(entityIsWalking(entity), "Entity is already walking");
  assertTrue(entityCanWalk(entity, entity->dir, NULL), "Entity cannot walk");

  int8_t tX, tY;
  entityDirGetPosition(entity->dir, 1, &tX, &tY);

  entity->y += tY;
  entity->x += tX;
  entity->subX = ENTITY_MOVE_SUBPIXEL * -tX;
  entity->subY = ENTITY_MOVE_SUBPIXEL * -tY;
}

bool_t entityIsWalking(const entity_t *entity) {
  assertNotNull(entity, "Entity cannot be NULL");
  return (entity->subX != 0 || entity->subY != 0);
}

bool_t entityCanWalk(
  const entity_t *entity,
  const entitydir_t dir,
  entity_t **entityInWay
) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertTrue(dir < ENTITY_DIR_COUNT, "Invalid entity direction");

  int8_t dX, dY;
  entityDirGetPosition(dir, 1, &dX, &dY);

  uint8_t tX = entity->x + dX;
  if(tX < 0 || tX >= MAP.width) return false;
  
  uint8_t tY = entity->y + dY;
  if(tY < 0 || tY >= MAP.height) return false;

  if(entityInWay == NULL) {
    if(mapGetEntityAt(tX, tY) != NULL) return false;
  } else {
    *entityInWay = mapGetEntityAt(tX, tY);
    if(*entityInWay != NULL) return false;
  }

  return true;
}

void entityPositionSet(entity_t *entity, const uint8_t x, const uint8_t y) {
  assertNotNull(entity, "Entity cannot be NULL");
  assertTrue(x < MAP.width, "X coordinate out of bounds");
  assertTrue(y < MAP.height, "Y coordinate out of bounds");
  entity->x = x;
  entity->y = y;
  entity->subX = 0;
  entity->subY = 0;
}

bool_t entityInteract(entity_t *interacted, entity_t *player) {
  assertNotNull(interacted, "Interacted entity cannot be NULL");
  assertNotNull(player, "Player entity cannot be NULL");
  assertTrue(interacted->type != ENTITY_TYPE_NULL, "Interacted entity invalid");
  assertTrue(player->type == ENTITY_TYPE_PLAYER, "Entity is not a player");

  if(ENTITY_CALLBACKS[interacted->type].interact != NULL) {
    return ENTITY_CALLBACKS[interacted->type].interact(interacted, player);
  }

  return false;
}