/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "player.h"
#include "npc.h"

#define ENTITY_WIDTH 16
#define ENTITY_HEIGHT 16
#define ENTITY_MOVE_SUBPIXEL ENTITY_WIDTH

typedef enum {
  ENTITY_DIR_UP = 0,
  ENTITY_DIR_DOWN = 1,
  ENTITY_DIR_LEFT = 2,
  ENTITY_DIR_RIGHT = 3,

  ENTITY_DIR_NORTH = ENTITY_DIR_UP,
  ENTITY_DIR_SOUTH = ENTITY_DIR_DOWN,
  ENTITY_DIR_WEST = ENTITY_DIR_LEFT,
  ENTITY_DIR_EAST = ENTITY_DIR_RIGHT
} entitydir_t;

#define ENTITY_DIR_COUNT (ENTITY_DIR_RIGHT + 1)

typedef enum {
  ENTITY_TYPE_NULL = 0,
  ENTITY_TYPE_PLAYER = 1,
  ENTITY_TYPE_NPC = 2,
} entitytype_t;
#define ENTITY_TYPE_COUNT (ENTITY_TYPE_NPC + 1)

typedef struct _entity_t {
  entitytype_t type;
  uint8_t x, y;
  int8_t subX, subY;
  entitydir_t dir;

  // Per type data
  union {
    player_t player;
    npc_t npc;
  };
} entity_t;

typedef struct {
  void (*init)(entity_t *entity);
  void (*update)(entity_t *entity);
  bool_t (*interact)(entity_t *self, entity_t *player);
} entitycallbacks_t;
extern entitycallbacks_t ENTITY_CALLBACKS[ENTITY_TYPE_COUNT];

/**
 * Initializes an entity with the given type.
 * 
 * @param entity Pointer to the entity to initialize.
 * @param type The type of the entity to initialize.
 */
void entityInit(entity_t *entity, const entitytype_t type);

/**
 * Updates the entity's state based on its type.
 * 
 * @param entity Pointer to the entity to update.
 */
void entityUpdate(entity_t *entity);

/**
 * Turns the entity to face a specific direction.
 * 
 * @param entity Pointer to the entity to turn.
 * @param dir The direction to turn the entity towards.
 */
void entityTurn(entity_t *entity, const entitydir_t dir);

/**
 * Gets the position of a specific direction at a given distance.
 * 
 * @param dir The direction to get the position in.
 * @param distance The distance to move in that direction.
 * @param outX Pointer to store the resulting x-coordinate.
 * @param outY Pointer to store the resulting y-coordinate.
 */
void entityDirGetPosition(
  const entitydir_t dir,
  const uint8_t distance,
  int8_t *outX,
  int8_t *outY
);

/**
 * Gets the look direction for a given entity to be looking at the specified
 * coordinates.
 * 
 * @param entity Pointer to the entity to get the look direction for.
 * @param x The x-coordinate to look at.
 * @param Y The y-coordinate to look at.
 * @return The direction the entity should look towards.
 */
entitydir_t entityGetLookDirection(
  const entity_t *entity,
  const uint8_t x,
  const uint8_t Y
);

/**
 * Makes the entity walk in the current direction.
 * 
 * @param entity Pointer to the entity to make walk.
 */
void entityWalk(entity_t *entity);

/**
 * Checks if the entity is currently mid-walking.
 * 
 * @param entity Pointer to the entity to check.
 * @return true if the entity is walking, false otherwise.
 */
bool_t entityIsWalking(const entity_t *entity);

/**
 * Checks if the entity can walk in a specific direction.
 * 
 * @param entity Pointer to the entity to check.
 * @param dir The direction to check for walking capability.
 * @param inWay Pointer to store any entity in the way, if applicable.
 * @return true if the entity can walk in the specified direction, false otherwise.
 */
bool_t entityCanWalk(
  const entity_t *entity,
  const entitydir_t dir,
  entity_t **inWay
);

/**
 * Sets the position of the entity.
 * 
 * @param entity Pointer to the entity to set the position for.
 * @param x The x-coordinate to set.
 * @param y The y-coordinate to set.
 */
void entityPositionSet(entity_t *entity, const uint8_t x, const uint8_t y);

/**
 * Handles interaction between an entity and a player.
 * 
 * @param self Pointer to the entity that is being interacted with.
 * @param player Pointer to the player entity interacting with the entity.
 * @return true if interaction happened.
 */
bool_t entityInteract(
  entity_t *self,
  entity_t *player
);